import json
import re
import logging
from collections import namedtuple
from typing import NamedTuple, Tuple, Union

from . import client, exceptions

logger = logging.getLogger(__name__)

_ping_header = ["Succeeded", "Address", "Resolved Address", "Round Trip", "Errors"]
_ping_fields = ["succeeded", "address", "resolved_address", "round_trip", "errors"]
_ping = namedtuple("ping", _ping_fields)  # type: ignore

_dns_lookup_header = ["Hostname", "Resolved Address", "Succeeded", "Errors"]
_dns_lookup_fields = ["hostname", "resolved_address", "succeeded", "errors"]
_dns_lookup = namedtuple("dns_lookup", _dns_lookup_fields)  # type: ignore


def ping(connection, *, address: str) -> NamedTuple:
    """ Example list from the powershell cmdlet:

        Status        : Success
        Address       : 212.58.249.213
        RoundtripTime : 4

        This doesn't use the 'better' Test-Connection command
        because that randomly fails with resource errors -_-.
    """
    try:
        output = client.run_powershell(
            connection,
            script=(
                f"""
$ping = New-Object System.Net.NetworkInformation.Ping
$ping.send('{address}') | Select Status, Address, RoundTripTime | Format-List"""
            ),
        )
        logger.info(f"ping output: {output}")

    except exceptions.PowershellError as e:
        if "An exception occurred during a Ping request." in str(e):
            # I have only seen this happen when using a DNS name that doesn't resolve.
            # Hard to do any better than this because the error reporting is so bad.
            return _ping(  # type: ignore
                succeeded=False,
                address=address,
                resolved_address=None,
                round_trip=None,
                errors=(f"Powershell Exception. If '{address}' is a " "hostname, it may not have resolved."),
            )
        raise e

    # Example output: "Status        : Success\nAddress       : 212.58.249.211\nRoundtripTime : 3"
    regex = re.compile(
        (
            r"\s*Status\s+\:\s(?P<status>(\w+))\n\s*"
            r"Address\s+\:\s*(?P<address>(.*))\n\s*"
            r"RoundtripTime\s+\:\s(?P<round_trip>(\w+)).*"
        )
    )
    named_groups = re.search(regex, output).groupdict()
    expected_keys = {"status", "address", "round_trip"}
    if not all(list(map(lambda key: key in named_groups, expected_keys))):
        raise Exception(
            f"Could not extract all keys {expected_keys} "
            "from output when generating ping result. Raw "
            f"output: '{repr(output)}'. Parsed output: {named_groups}"
        )

    def _resolve_status(status) -> Tuple[bool, Union[str, None]]:
        # Take a status string from powershell and
        # return a tuple of resolved values:
        #
        # (succeeded, error_message)
        #
        # If an unhandled status is provided, default
        # to failure.
        status = status.lower().strip()
        try:
            return {
                "success": (True, None),
                "timedout": (False, f"Address '{address}' timed out."),
                "destinationhostunreachable": (False, f"Host '{address}' unreachable."),
            }[status]
        except KeyError:
            return (False, f"Unknown ping status '{status}'.")

    succeeded, errors = _resolve_status(named_groups["status"])

    return _ping(  # type: ignore
        succeeded=succeeded,
        address=address,
        resolved_address=named_groups["address"],
        round_trip=named_groups["round_trip"],
        errors=errors,
    )


def dns_lookup(connection, *, hostname: str):
    script = f"Resolve-DNSName -Name {hostname} | Select IP4Address | ConvertTo-Json"
    try:
        output = client.run_powershell(connection, script=script)
        logger.info(f"dns lookup output: {output}")
    except exceptions.PowershellError as e:
        if "name does not exist" in str(e):
            return _dns_lookup(  # type: ignore
                hostname=hostname, succeeded=False, resolved_address=None, errors=f"Name '{hostname}' does not exist."
            )
        raise e

    if output:
        loaded_results = json.loads(output)
        resolved_addresses = [res["IP4Address"] for res in loaded_results if res.get("IP4Address")]
        # strip out intermediate DNS forwarding results
        return _dns_lookup(  # type: ignore
            hostname=hostname, succeeded=True, resolved_address=resolved_addresses[-1], errors=None
        )
    else:
        raise Exception("DNS resolution did not return an error, but the response is blank.")
