"""
Domain

This module contains actions for related to Active Directory domains.

!!! warning "Domain Servers"

    The actions in this module will only work on domain joined servers.

    If an action requires a Powershell module to be installed, its documentation will let you know.
"""
import logging

from ppa import ui
from windows_server import client

from .records import WindowsServer
from ._wrappers import winrm


logger = logging.getLogger(__name__)


class NoNetBiosName(Exception):
    pass


@winrm
def get_netbios_domain_name(windows_server: WindowsServer) -> str:
    r"""Get the NetBIOS name of the Active Directory domain the server is joined to.

The NetBIOS domain name is the short domain used in the _domain\username_ format.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

??? success "Output"

    The NetBIOS domain name.

??? example

    ```YAML linenums="1"
    - windows_server.domain.get_netbios_domain_name:
      load:
        windows_server: windows_server
      save: netbios_domain_name
    ```
"""

    # 2012-DOMAIN    <00>  GROUP       Registered

    try:
        return client.run_powershell(
            windows_server,
            script=(
                """$ADSystemInfo = New-Object -ComObject "ADSystemInfo"
$ADSystemInfo.GetType().InvokeMember("DomainShortName", "GetProperty", $null, $ADSystemInfo, $null)"""
            ),
        )
    except Exception as e:
        logger.exception(e)
        ui.output_error("Could not identify NetBIOS domain name on the server. Please view the logs for more information.")
        raise NoNetBiosName("Could not identify NetBIOS domain name on the server.")
